#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <hpic.h>

int main(int argc, char *argv[])
{
  char ch = '-';
  size_t i, j, k;

  char infile[HPIC_STRNL];
  char outfile[HPIC_STRNL];
  int makevec = 0;
  int makevecindx = 0;
  int makefull = 0;
  int makecut = 0;
  int checkme = 0;

  size_t nside;
  int order = 0;
  int coord = HPIC_COORD_C;
  int tabtype = 0;
  size_t nmaps;
  size_t nvecs;
  size_t ncol;
  size_t length;
  size_t nkeys;
  int skip;

  char comment[HPIC_STRNL];
  char creator[HPIC_STRNL];
  char extname[HPIC_STRNL];
  char units[HPIC_STRNL];
  char line[HPIC_STRNL];
  size_t linelen;
  char *words[5];

  hpic_vec_int *indx = NULL;
  char **vecnames = NULL;
  char **vecunits = NULL;
  hpic_vec_fltarr *vecs = NULL;
  hpic_vec_float *tempvec;

  hpic_int *pix = NULL;
  hpic_fltarr *maps = NULL;
  hpic_int *hits = NULL;
  hpic_float *errs = NULL;
  hpic_keys *keys = NULL;
  hpic_float *tempmap;
  int keyint;
  float keyfloat;

  float ftemp;
  int itemp;
  int temp;
  char tempstr[HPIC_STRNL];
  FILE *fp;

  order = HPIC_NEST;
  coord = HPIC_COORD_C;
  nside = 0;
  strcpy(creator, "hpic_2fits");

  if (argc == 1) {
    printf("\n");
    printf("<******************************************************>\n");
    printf("Usage:  hpic_2fits\n");
    printf("        -out <output FITS file>\n");
    printf("        -vec (build vector FITS file)\n");
    printf("        -vecindx (build indexed vector FITS file)\n");
    printf("        -full (build full-sphere map FITS file)\n");
    printf("        -cut (build cut-sphere map FITS file)\n");
    printf("        -nside <nside of maps>\n");
    printf
      ("        -order <0=NESTED 1=RING> (only applies to map FITS files)\n");
    printf
      ("        -table <0=BINARY 1=ASCII> (only applies to vector FITS files)\n");
    printf("        -units <units>\n");
    printf("        -coord <coordsys (eg. C, G, O)>\n");
    printf("        <input text file (properly formatted)>\n");
    printf("<******************************************************>\n\n");
    exit(1);
  }

  /* parse command line */

  strncpy(infile, argv[argc - 1], HPIC_STRNL);
  for (i = 0; i < argc; i++) {
    if (argv[i][0] == ch) {
      if (!strcmp(argv[i], "-out")) {
        strncpy(outfile, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-units")) {
        strncpy(units, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-nside")) {
        nside = (size_t) atoi(argv[i + 1]);
        i++;
      }
      if (!strcmp(argv[i], "-vec")) {
        makevec = 1;
        checkme++;
      }
      if (!strcmp(argv[i], "-vecindx")) {
        makevecindx = 1;
        checkme++;
      }
      if (!strcmp(argv[i], "-full")) {
        makefull = 1;
        checkme++;
      }
      if (!strcmp(argv[i], "-cut")) {
        makecut = 1;
        checkme++;
      }
      if (!strcmp(argv[i], "-order")) {
        order = atoi(argv[i + 1]);
        if (order == 0) {
          order = HPIC_NEST;
        } else {
          order = HPIC_RING;
        }
        i++;
      }
      if (!strcmp(argv[i], "-table")) {
        tabtype = atoi(argv[i + 1]);
        if (tabtype == 0) {
          tabtype = HPIC_FITS_BIN;
        } else {
          tabtype = HPIC_FITS_ASCII;
        }
        i++;
      }
      if (!strcmp(argv[i], "-coord")) {
        if (argv[i + 1][0] == 'C') {
          coord = HPIC_COORD_C;
        } else if (argv[i + 1][0] == 'G') {
          coord = HPIC_COORD_G;
        } else if (argv[i + 1][0] == 'E') {
          coord = HPIC_COORD_E;
        } else if (argv[i + 1][0] == 'O') {
          coord = HPIC_COORD_O;
        }
        i++;
      }
    }
  }

  if ((checkme < 1) || (checkme > 1)) {
    fprintf(stderr,
            "hpic_2fits:  you must specify exactly one type of FITS file\n");
    exit(1);
  }

  /* read in text file and create FITS file */

  if ((fp = fopen(infile, "r")) == NULL) {
    fprintf(stderr, "hpic_2fits:  Cannot open text file %s!\n", infile);
    exit(1);
  }

  fgets(line, HPIC_STRNL, fp);
  linelen = strlen(line);
  if ((0 < linelen) && ('\n' == line[linelen - 1])) {
    line[linelen - 1] = '\0';
  }
  k = 0;
  j = 0;
  while ((k < 4) && (line[j] != '\0')) {
    while ((line[j] != '\0') && (line[j] == ' ')) {
      j++;
    }
    if (line[j] != '\0') {
      words[k] = &(line[j]);
      k++;
    }
    while ((line[j] != '\0') && (line[j] != ' ')) {
      j++;
    }
    if (k < 4) {
      line[j] = '\0';
    }
    j++;
  }
  nkeys = (size_t) atoi(words[1]);
  ncol = (size_t) atoi(words[2]);
  length = (size_t) atoi(words[3]);

  if (makefull && (length != 12 * nside * nside)) {
    fprintf(stderr,
            "hpic_2fits:  for full-sphere FITS file, you must specify 12 * nside^2 rows!\n");
    exit(1);
  }
  keys = hpic_keys_alloc();
  for (i = 0; i < nkeys; i++) {
    fgets(line, HPIC_STRNL, fp);
    linelen = strlen(line);
    if ((0 < linelen) && ('\n' == line[linelen - 1])) {
      line[linelen - 1] = '\0';
    }
    k = 0;
    j = 0;
    while ((k < 5) && (line[j] != '\0')) {
      while ((line[j] != '\0') && (line[j] == ' ')) {
        j++;
      }
      if (line[j] != '\0') {
        words[k] = &(line[j]);
        k++;
      }
      while ((line[j] != '\0') && (line[j] != ' ')) {
        j++;
      }
      if (k < 5) {
        line[j] = '\0';
      }
      j++;
    }
    if (strncmp(words[1], "SKEY", HPIC_STRNL) == 0) {
      hpic_keys_sadd(keys, words[2], words[3], words[4]);
    } else if (strncmp(words[1], "IKEY", HPIC_STRNL) == 0) {
      keyint = atoi(words[3]);
      hpic_keys_iadd(keys, words[2], keyint, words[4]);
    } else if (strncmp(words[1], "FKEY", HPIC_STRNL) == 0) {
      keyfloat = (float)atof(words[3]);
      hpic_keys_fadd(keys, words[2], keyfloat, words[4]);
    }
  }

  if (makevec) {
    nvecs = ncol;
    vecs = hpic_vec_fltarr_alloc(nvecs);
    vecnames = hpic_strarr_alloc(nvecs);
    vecunits = hpic_strarr_alloc(nvecs);
    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_float_alloc(length);
      hpic_vec_fltarr_set(vecs, i, tempvec);
      sprintf(vecnames[i], "Vector %lu", i);
      strcpy(vecunits[i], "Unknown");
    }
    for (i = 0; i < length; i++) {
      for (j = 0; j < nvecs; j++) {
        fscanf(fp, "%f", &ftemp);
        tempvec = hpic_vec_fltarr_get(vecs, j);
        hpic_vec_float_set(tempvec, i, ftemp);
      }
      fscanf(fp, "\n");
    }
    snprintf(comment, HPIC_STRNL, "made from input file %s", infile);
    strcpy(extname, "Vector Table");
    hpic_fits_vec_write(outfile, tabtype, creator, extname, comment, vecs,
                        vecnames, vecunits, keys);
    hpic_strarr_free(vecnames, nvecs);
    hpic_strarr_free(vecunits, nvecs);
    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_fltarr_get(vecs, i);
      hpic_vec_float_free(tempvec);
    }
    hpic_vec_fltarr_free(vecs);
  }

  if (makevecindx) {
    if (ncol < 2) {
      fprintf(stderr,
              "hpic_2fits:  Cannot make an indexed vector file with < 2 columns!\n");
      exit(1);
    }
    if (tabtype != HPIC_FITS_BIN) {
      fprintf(stderr,
              "hpic_2fits:  vector index files must be a BINARY table, overriding user value\n");
    }
    nvecs = ncol - 1;
    vecs = hpic_vec_fltarr_alloc(nvecs);
    vecnames = hpic_strarr_alloc(nvecs);
    vecunits = hpic_strarr_alloc(nvecs);
    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_float_alloc(length);
      hpic_vec_fltarr_set(vecs, i, tempvec);
      sprintf(vecnames[i], "Vector %lu", i);
      strcpy(vecunits[i], "Unknown");
    }

    indx = hpic_vec_int_alloc(length);
    for (i = 0; i < length; i++) {
      fscanf(fp, "%d", &itemp);
      hpic_vec_int_set(indx, i, itemp);
      for (j = 0; j < nvecs; j++) {
        fscanf(fp, "%f", &ftemp);
        tempvec = hpic_vec_fltarr_get(vecs, j);
        hpic_vec_float_set(tempvec, i, ftemp);
      }
      fscanf(fp, "\n");
    }
    snprintf(comment, HPIC_STRNL, "made from input file %s", infile);
    strcpy(extname, "Indexed Vector Table");
    hpic_fits_vecindx_write(outfile, creator, extname, comment, indx, vecs,
                            vecnames, vecunits, keys);
    hpic_strarr_free(vecnames, nvecs);
    hpic_strarr_free(vecunits, nvecs);
    hpic_vec_int_free(indx);
    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_fltarr_get(vecs, i);
      hpic_vec_float_free(tempvec);
    }
    hpic_vec_fltarr_free(vecs);
  }

  if (makefull) {
    nmaps = ncol;
    maps = hpic_fltarr_alloc(nmaps);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      sprintf(tempstr, "Map %lu", i);
      hpic_float_name_set(tempmap, tempstr);
      hpic_float_units_set(tempmap, units);
      hpic_fltarr_set(maps, i, tempmap);
    }
    for (i = 0; i < length; i++) {
      for (j = 0; j < nmaps; j++) {
        fscanf(fp, "%f", &ftemp);
        tempmap = hpic_fltarr_get(maps, j);
        hpic_float_set(tempmap, i, ftemp);
      }
      fscanf(fp, "\n");
    }
    snprintf(comment, HPIC_STRNL, "made from input file %s", infile);
    strcpy(extname, "Full-Sphere Maps");
    hpic_fits_full_write(outfile, creator, extname, comment, maps, keys);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_fltarr_get(maps, i);
      hpic_float_free(tempmap);
    }
    hpic_fltarr_free(maps);
  }

  if (makecut) {
    if (ncol < 4) {
      fprintf(stderr,
              "hpic_2fits:  Cannot make an cut-sphere file with < 4 columns!\n");
      exit(1);
    }
    nmaps = ncol - 3;
    pix = hpic_int_alloc(nside, order, coord, HPIC_AUTO);
    hpic_int_name_set(pix, "PIXEL");
    hits = hpic_int_alloc(nside, order, coord, HPIC_AUTO);
    hpic_int_name_set(hits, "HITS");
    errs = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
    hpic_float_name_set(errs, "ERROR");
    hpic_float_units_set(errs, "Unknown");
    maps = hpic_fltarr_alloc(nmaps);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      sprintf(tempstr, "Map %lu", i);
      hpic_float_name_set(tempmap, tempstr);
      hpic_float_units_set(tempmap, units);
      hpic_fltarr_set(maps, i, tempmap);
    }
    for (i = 0; i < length; i++) {
      fscanf(fp, "%d", &itemp);
      hpic_int_set(pix, (size_t) itemp, itemp);
      for (j = 0; j < nmaps; j++) {
        fscanf(fp, "%f", &ftemp);
        tempmap = hpic_fltarr_get(maps, j);
        hpic_float_set(tempmap, (size_t) itemp, ftemp);
      }
      fscanf(fp, "%d", &temp);
      hpic_int_set(hits, (size_t) itemp, temp);
      fscanf(fp, "%f", &ftemp);
      hpic_float_set(errs, (size_t) itemp, ftemp);
      fscanf(fp, "\n");
    }
    snprintf(comment, HPIC_STRNL, "made from input file %s", infile);
    strcpy(extname, "Cut-Sphere Maps");
    hpic_fits_cut_write(outfile, creator, extname, comment, pix, hits, errs,
                        maps, keys);
    hpic_float_free(errs);
    hpic_int_free(hits);
    hpic_int_free(pix);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_fltarr_get(maps, i);
      hpic_float_free(tempmap);
    }
    hpic_fltarr_free(maps);
  }

  hpic_keys_free(keys);
  fclose(fp);

  return 0;
}
